/*****************************************************************************
 **                      TAU Portable Profiling Package                     **
 **                      http://www.cs.uoregon.edu/research/paracomp/tau    **
 *****************************************************************************
 ** 	File: 		MasterMind_CCA.cc                                   **
 ** 	Description: 	MasterMind component, generates output files        **
 **                     that track parameter values to instrumented         **
 **			component ports                                     **
 ** 	Author: 	Nick Trebon?                                        **
 *****************************************************************************
 */

/*
 *  Changes 07-01-2004 - Alan Morris
 *    Fixed segfaults in destructor when component is created, but not used
 */

#include "cca.h"
#include "MasterMind_CCA.h"

#include <iostream>
#include <stdlib.h> // for getenv


namespace monitor {
  namespace ccaimpl {

    /** 
     * Default constructor
     */

    MasterMind::MasterMind () {
      //std::cout << "Begin MasterMind constructor\n";
      init_mport=false;
      init_tree = false;
      frameworkServices = 0;
      measurement_m=0;
      call_path_tree = 0;
      /*
      char *dirname;
      if ((dirname = getenv("MASTERMINDDIR")) == NULL) {
	// Use default directory name .
	dirname  = new char[8];
	strcpy (dirname,".");
      }
      */
      //      std::cout << "End MasterMind constructor\n";

    }
    
    MasterMind::~MasterMind () {
      //std::cout << "Begin MasterMind destructor\n";
      
      // delete records

      int rank = 0;
 
      if (records.empty() == false) {
	rank = (*records.begin()->second).getRank();
      }

      
      /*
      //temp
      char buffer[4096];
      FILE *f = popen("/bin/ls","r");
      while (fscanf(f,"%s",buffer) != EOF) {
	printf ("%s, ",buffer);
      }
      */

      map<string,performance::Record*>::iterator r;
      r = records.begin();
      
      // delete each record, also generate an entry in the 
      // database file for each record
      char fname[256];


      char *dirname;
      if ((dirname = getenv("MASTERMINDDIR")) == NULL) {
	// Use default directory name .
	dirname  = new char[8];
	strcpy (dirname,".");
      }

      string dbfilename = "dbfile";
      sprintf(fname, "%s/%s.%d", dirname, dbfilename.c_str(), rank);
      ofstream outfile(fname, std::ios::app);
      while (r!=records.end()) {
	if (r->second) {
	  outfile << r->first << "\t" << (r->second)->getFileName() << endl;
	  delete r->second;
	}
	r++;
      }
    
      outfile.close();
      records.clear();
     
      if (init_tree) {
	//delete call path tree
	string fn = "call_path_tree";
	char ctname[256];

	
	sprintf(ctname, "%s/%s.%d", dirname, fn.c_str(), rank);
	
	std::cout << "writing to " << ctname << endl;
	call_path_tree->dumpChildren((string) ctname);
	if (call_path_tree)
	  delete call_path_tree;
      }
      //      std::cout<<"End MasterMind destructor\n";
    }

    /**
     * Create a record and store it in the records map
     */
    void MasterMind::createRecord(string rname) {
      // first check to see if we have gotten a measurement port and MPIBorrow port
      // Only do once!!

      if (!init_mport) {
	//Get Measurement port
	classic::gov::cca::Port * port=frameworkServices->getPort("MeasurementPort");
	if(port)
	  measurement_m = dynamic_cast < performance::ccaports::Measurement * >(port);
	if(measurement_m==0) {
	  std::cerr << "Connected to something other than a Measurement port" << std::endl;
	}

	init_mport=true;
      }

      // if no record with that name exists, create it.  Otherwise, do nothing.
      if (records.find(rname) == records.end()){
	performance::Record *r = new performance::Record(rname, frameworkServices, measurement_m);
	// check to see if we are using MPI.  To do this, see if MPI_INIT has been   
	// called.  If it hasn't set rank to 0 for single cpu case.  Else, get   
	// a communicator and determine the cpu's rank.  
	int flag;
	MPI_Initialized(&flag);
	if (!flag) {
	  r->setRank(0);
	} else {
	  classic::gov::cca::MPIBorrow * pMPI_Port = dynamic_cast<classic::gov::cca::MPIBorrow *> (frameworkServices->getPort("mpi"));
	  // otherwise, get the communicator and set the rank 
	  int key,tag,rank;
	  MPI_Comm comm = pMPI_Port->borrowComm(1, &tag, key);
	  MPI_Comm_rank(comm, &rank);
	  r->setRank(rank);
	}
	r->setFileName(rname);
	records[rname]=r;
      }
    }

    /**
     * Start monitoring the given record
     **/    
    void MasterMind::startMonitoring(string rname, map<string,double> &m) {
      //      std::cout << "MasterMind::startMonitoring here!\n";
      //first check to see if we have initialized the call path tree
      //if we haven't, do so, otherwise, we need to advance
      string node_name = rname.substr(0,rname.find("::"));
      if (!init_tree) {
	call_path_tree = new CallPathNode(node_name);
	init_tree = true;
      } else {
	call_path_tree = call_path_tree->getChild(node_name);
      }
      
      (*records[rname]).setParameters(m);
      (*records[rname]).startTiming(true);
    }
    
    /**
     * Stop monitoring the given record
     **/
    void MasterMind::stopMonitoring(string rname) {
      map<string,double> counters;
      (*records[rname]).stopTiming(true, counters);
      //when we stop monitoring, assume that compute in component is over, so 
      //we can move back "up" the call-path tree (first, increment the performance 
      //counters for the current node)
      call_path_tree->incrementCounters(counters);
      //before moving back up tree, make sure we have a valid pointer for the root
      //if we don't assume we are at the root.
      if (call_path_tree->getParent())
	call_path_tree = call_path_tree->getParent();
    }

    /**
     * Set dump file name
     **/
    void MasterMind::setFileName(string rname, string fname) {
      (*records[rname]).setFileName(fname);
    }

    /**
     * Dump data for the given record
     **/
    void MasterMind::dumpData(string rname) {
      (*records[rname]).dumpToFile(rname);
    }

    /**
     * Dump data for the given record to the given file
     **/
    void MasterMind::dumpData(string rname, string fname) {
      (*records[rname]).setFileName(fname);
      (*records[rname]).dumpToFile(rname);
    }

    /**
     * Destroy the given record
     **/
    void MasterMind::destroyRecord(string rname) {
      map<string,  performance::Record *>::iterator it;
      it = records.find(rname);
      delete it->second;
      records.erase(it);    
    }

    /**
     * The framework passes a pointer to the Services object
     * via this method, which is called by the framework as soon
     * as the component is instantiated.
     */
    void MasterMind::setServices (classic::gov::cca::Services * services) {

      if (services != 0) {
	frameworkServices = services;
	
	// Provide a Monitor port
	classic::gov::cca::PortInfo * portInfo =
	  frameworkServices->createPortInfo ("MonitorPort",
					     "monitor.ccaports.Monitor",
					     0);
	if (portInfo != 0)
	  frameworkServices->addProvidesPort (this, portInfo);
	
	// Use a Measurement port
	portInfo = 
	  frameworkServices->createPortInfo ("MeasurementPort",
					     "performance.ccaports.Measurement",
					     0);
	frameworkServices->registerUsesPort(portInfo);
	
	// Use a MPIBorrow port
	frameworkServices->registerUsesPort( frameworkServices->createPortInfo("mpi", "gov.cca.MPIBorrow",0));
      } else {
	// Close down if not closed already
	if (frameworkServices){
	  frameworkServices->releasePort ("MeasurementPort");
	  frameworkServices->releasePort("mpi");
	  frameworkServices->removeProvidesPort ("MonitorPort");
	}
	frameworkServices = 0;
      }
    }
  }
}
